# MIT License
#
# Copyright (c) 2017 Laboratory for Computational and Statistical Learning
#
# authors: Daniele Calandriello
# email:   daniele.calandriello@iit.it
# Website: http://lcsl.mit.edu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dppy.bless import bless, CentersDictionary, estimate_rls_bless
from dppy.utils import stable_invert_root, evaluate_L_diagonal, get_progress_bar, check_random_state
import numpy as np
from scipy.optimize import brentq
from collections import namedtuple


def compute_nystrom_dict(X_data, eval_L, rls_oversample_bless, rls_oversample_dppvfx, rng, nb_iter_bless=None, verbose=True):
    """ Computes the initial dictionary necessary for the algorithm. Internally invoke BLESS.

    :param array_like X_data: dataset that we must approximate
    :param callable eval_L: likelihood function
    :param float rls_oversample_bless: see :func:`vfx_sampling_precompute_constants`
    :param float rls_oversample_dppvfx: see :func:`vfx_sampling_precompute_constants`
    :param RandomState rng: random source used for sampling
    :param int nb_iter_bless:  iterations for BLESS, if None it is set to log(n)
    :param bool verbose: controls verbosity of debug output, including progress bars.
        the progress bar reports:
        - lam: lambda value of the current iteration
        - m: current size of the dictionary (number of centers contained)
        - m_expected: expected size of the dictionary before sampling
        - probs_dist: (mean, max, min) of the approximate rlss at the current iteration
    :return: an (eps, lambda)-accurate dictionary for Nystrom approximation
    :rtype:
        CentersDictionary

    """
    n, _ = X_data.shape

    # Phase 1: compute initial dictionary D_bless with small rls_oversample_bless
    # D_bless is used only to estimate all RLS

    dict_bless = bless(X_data, eval_L, 1.0, rls_oversample_bless, rng, nb_iter_bless=nb_iter_bless, verbose=verbose)

    bless_rls_estimate = estimate_rls_bless(dict_bless, X_data, eval_L, 1.0)

    # Phase 2: use estimate RLS to sample the dict_dppvfx dictionary, i.e. the one used to construct A
    # here theory says that to have high acceptance probability we need the oversampling factor to be ~deff^2
    # but even with constant oversampling factor we seem to accept fast

    probs = np.minimum(rls_oversample_dppvfx * bless_rls_estimate, 1.0)
    if not np.all(probs >= 0.0):
        raise ValueError('Some estimated RLS is negative, this should never happen. Min prob: {}'.format(np.min(probs)))

    selected = rng.rand(n) <= probs
    s = selected.sum()

    if not s > 0:
        raise ValueError('No point selected during RLS sampling step, try to increase rls_oversample_bless. '
                         'Expected number of points: {:.3f}'.format(probs.sum()))

    dict_dppvfx = CentersDictionary(idx=selected.nonzero()[0],
                                    X=X_data[selected, :],
                                    probs=probs[selected],
                                    lam=1,
                                    rls_oversample=rls_oversample_dppvfx)

    return dict_dppvfx


def estimate_rls_from_embedded_points(eigvec, eigvals, B_bar_T, diag_L, diag_L_hat, alpha_star):
    """ Given embedded points, and a decomposition of embedded covariance matrix, estimate RLS.
    Note that this is a different estimator than the one used in BLESS (i.e. :func:`dppy.bless.estimate_rls_bless`),
    which we use here for efficiency because we can recycle already embedded points and eigen-decomposition.

    :param array_like eigvec: eigenvectors of I_A_mm = B_bar_T*B_bar_T.T + lam I, see :func:`vfx_sampling_precompute_constants`
    :param array_like eigvals: eigenvalues of I_A_mm = B_bar_T*B_bar_T.T + lam I, see :func:`vfx_sampling_precompute_constants`
    :param array_like B_bar_T: (m x n) transposed matrix of n points embedded using a dictionary with m centers
    :param array_like diag_L: diagonal of L
    :param array_like diag_L_hat: diagonal of L_hat, the Nystrom approximation of L
    :param float alpha_star: a rescaling factor used to adjust the expected size of the DPP sample
    :return: RLS estimates for all rows in B_bar_T
    :rtype:
        array_like
    """

    U_DD, S_root_inv_DD = stable_invert_root(eigvec, np.maximum(eigvals, 0))

    E = S_root_inv_DD * U_DD.T

    X_precond = E.dot(B_bar_T)

    rls_estimate =\
        diag_L * alpha_star\
        - diag_L_hat * alpha_star\
        + np.square(X_precond, out=X_precond).sum(axis=0)

    return rls_estimate


_IntermediateSampleInfo =\
    namedtuple('_IntermediateSampleInfo',
               ['alpha_star', 'logdet_I_A', 'q', 's', 'z', 'rls_estimate'])


def vfx_sampling_precompute_constants(X_data,
                                      eval_L,
                                      rng,
                                      desired_expected_size=None,
                                      rls_oversample_dppvfx=4.0,
                                      rls_oversample_bless=4.0,
                                      nb_iter_bless=None,
                                      verbose=True):
    """Pre-compute quantities necessary for the vfx rejection sampling loop, such as the inner Nystrom approximation,
    and the RLS of all elements in L.

        :param array_like X_data: dataset such that L = eval_L(X_data), out of which we aresampling objects according
        to a DPP
        :param callable eval_L: likelihood function. Given two sets of n points X and m points Y, eval_L(X, Y) should
        compute the (n x m) matrix containing the likelihood between points. The function should also
        accept a single argument X and return eval_L(X) = eval_L(X, X).
        As an example, see the implementation of any of the kernels provided by scikit-learn
        (e.g. sklearn.gaussian_process.kernels.PairwiseKernel).
        :param RandomState rng: random source used for sampling
        :param desired_expected_size: desired expected sample size for the DPP. If None, use the natural DPP expected
        sample size. The vfx sampling algorithm can approximately adjust the expected sample size of the DPP by
        rescaling the L matrix with a scalar alpha_star <= 1. Adjusting the expected sample size can be useful to
        control downstream complexity, and it is necessary to improve the probability of drawing a sample with
        exactly k elements when using vfx for k-DPP sampling. Currently only reducing the sample size is supported,
        and the sampler will return an exception if the DPP sample has already a natural expected size
        smaller than desired_expected_size.
        :type desired_expected_size:
            float or None, default None
        :param rls_oversample_dppvfx: Oversampling parameter used to construct dppvfx's internal Nystrom approximation.
        The rls_oversample_dppvfx >= 1 parameter is used to increase the rank of the approximation by
        a rls_oversample_dppvfx factor. This makes each rejection round slower and more memory intensive,
        but reduces variance and the number of rounds of rejections, so the actual runtime might increase or decrease.
        Empirically, a small factor rls_oversample_dppvfx = [2,10] seems to work. It is suggested to start with
        a small number and increase if the algorithm fails to terminate.
        :type rls_oversample_dppvfx:
            float, default 4.0
        :param rls_oversample_bless: Oversampling parameter used during bless's internal Nystrom approximation.
        Note that this is a different Nystrom approximation than the one related to :func:`rls_oversample_dppvfx`,
        and can be tuned separately.
        The rls_oversample_bless >= 1 parameter is used to increase the rank of the approximation by
        a rls_oversample_bless factor. This makes the one-time pre-processing slower and more memory intensive,
        but reduces variance and the number of rounds of rejections, so the actual runtime might increase or decrease.
        Empirically, a small factor rls_oversample_bless = [2,10] seems to work. It is suggested to start with
        a small number and increase if the algorithm fails to terminate or is not accurate.
        :type rls_oversample_bless:
            float, default 4.0
        :param int nb_iter_bless:  iterations for BLESS, if None it is set to log(n)
        :type nb_iter_bless:
            int or None, default None
        :param bool verbose: controls verbosity of debug output, including progress bars.
        The progress bar reports the inner execution of the bless algorithm, showing:
            - lam: lambda value of the current iteration
            - m: current size of the dictionary (number of centers contained)
            - m_expected: expected size of the dictionary before sampling
            - probs_dist: (mean, max, min) of the approximate rlss at the current iteration

        :return: Pre-computed information necessary for the vfx rejection sampling loop with fields
        - result.alpha_star: appropriate rescaling such that the expected sample size of DPP(alpha_star * L) is equal
        to a user-indicated constant desired_expected_size, or 1.0 if no such constant was specified by the user.
        - result.logdet_I_A: log determinant of the Nystrom approximation of L + I
        - result.q: placeholder q constant used for vfx sampling, to be replaced by the user before the sampling loop
        - result.s and result.z: approximations of the expected sample size of DPP(alpha_star * L) to be used in
        the sampling loop. For more details see [DeCaVa19]
        - result.rls_estimate: approximations of the RLS of all elements in X (i.e. in L)
        :rtype: _IntermediateSampleInfo

    """
    diag_L = evaluate_L_diagonal(eval_L, X_data)
    trace_L = diag_L.sum()

    # Phase 1: get all necessary prerequisite for preprocessing
    # construct D_A dictionary used to construct A and L_hat
    D_A = compute_nystrom_dict(X_data,
                               eval_L,
                               rls_oversample_bless,
                               rls_oversample_dppvfx,
                               rng,
                               nb_iter_bless=nb_iter_bless,
                               verbose=verbose)

    # Phase 2: pre-compute L_hat, B_bar, l_i, det(I + L_hat), etc.
    U_DD, S_DD, _ = np.linalg.svd(eval_L(D_A.X, D_A.X))
    U_DD, S_root_inv_DD = stable_invert_root(U_DD, S_DD)
    m = U_DD.shape[1]

    E = S_root_inv_DD * U_DD.T

    # The _T indicates that B_bar_T is the transpose of B_bar,
    # we keep it that way for efficiency reasons
    B_bar_T = E.dot(eval_L(D_A.X, X_data))
    diag_L_hat = np.square(B_bar_T).sum(axis=0)
    trace_L_hat = diag_L_hat.sum()

    # While we have L_hat = B_bar_T.T * B_bar_T, we do not want to compute explicitly the (n x n) matrix
    # instead we reason in terms of B_bar_T * B_bar_T.T which is a (m x m) matrix. We call this matrix A_mm.
    # I_A_mm indicates I + A_mm (i.e. A_mm with identity added)
    I_A_mm = B_bar_T.dot(B_bar_T.T)
    I_A_mm[np.diag_indices(m)] += 1.0

    # we now need to compute the l_i estimates using L_hat, it is more efficient to do it in terms of
    # B_bar_T and I_A_mm
    # in particular, we will use the diag(L - L_hat + L_hat(L_hat + I)^-1) estimator
    # but we must first tune L to obtain a desired s
    # we can use the fact the the non-zero eigenvalues of I + L_hat and I_A_mm are equal
    eigvals, eigvec = np.linalg.eigh(I_A_mm)

    if not np.all(eigvals >= 0.0):
        raise ValueError('Some eigenvalues of L_hat are negative, this should never happen. '
                         'Minimum eig: {}'.format(np.min(eigvals)))

    natural_expected_size = trace_L - trace_L_hat + np.sum((eigvals - 1.0) / eigvals)

    if not natural_expected_size >= 0.0:
        raise ValueError('natural_expected_size is negative, this should never happen. '
                         'natural_expected_size: {}'.format(natural_expected_size))

    # s might naturally be too large, but we can rescale L to shrink it
    # if we rescale alpha * L by a constant alpha,
    # s is now trace(alpha * L - alpha * L_hat + L_hat(L_hat + I / alpha)^-1)
    if desired_expected_size is None:
        alpha_star = 1.0
    elif natural_expected_size <= desired_expected_size:
        raise ValueError('The expected sample size is smaller than the desired sample size or k (if sampling from'
                         'a k-DPP).\n'
                         'This is unusual (i.e. you are trying to select more than the overall amount of diversity '
                         'in your set.\n'
                         'Increasing the expected sample size is currently not supported (only decreasing).\n'
                         'Please consider decreasing your k={} or changing L.'
                         ' Estimated mean cardinality: {}'.format(desired_expected_size,
                                                                  natural_expected_size))
    else:
        # since this is monotone in alpha, we can simply use Brent's algorithm (bisection + tricks)
        # it is a root finding algorithm so we must create a function with a root in desired_expected_size
        def temp_func_with_root_in_desired_expected_size(x):
            return (
                    x * trace_L
                    - x * trace_L_hat
                    + np.sum((eigvals - 1.0) / (eigvals - 1.0 + 1.0 / x))
                    - desired_expected_size
                   )

        alpha_star, opt_result = brentq(temp_func_with_root_in_desired_expected_size,
                                        a=10.0 * np.finfo(np.float).eps,
                                        b=1.0,
                                        full_output=True)

        if not opt_result.converged:
            raise ValueError('Could not find an appropriate rescaling for desired_expected_size.'
                             '(Flag, Iter, Root): {}'.format((opt_result.flag,
                                                              opt_result.iterations,
                                                              opt_result.root)))

    # adjust from I + A to I / alpha_star + A
    I_A_mm[np.diag_indices(m)] += 1.0 / alpha_star - 1.0
    eigvals += 1.0 / alpha_star - 1.0
    rls_estimate = estimate_rls_from_embedded_points(eigvec,
                                                     eigvals,
                                                     B_bar_T,
                                                     diag_L,
                                                     diag_L_hat,
                                                     alpha_star)

    if not np.all(rls_estimate >= 0.0):
        raise ValueError('Some estimate l_i is negative, this should never happen. '
                         'Minimum l_i: {}'.format(np.min(rls_estimate)))

    # s is simply the sum of l_i, and it is our proxy for the expected sample size. If desired_expected_size
    # is set, s should be very close to it
    s = np.sum(rls_estimate)
    if not s >= 0.0:
        raise ValueError('s is negative, this should never happen. '
                         's: {}'.format(s))

    # we need to compute z and logDet(I + L_hat)
    z = np.sum((eigvals - 1/alpha_star) / eigvals)

    # we need logdet(I + alpha * A) and we have eigvals(I / alpha_star + A) we can adjust using sum of logs
    logdet_I_A = np.sum(np.log(alpha_star * eigvals))

    if not logdet_I_A >= 0.0:
        raise ValueError('logdet_I_A is negative, this should never happen. '
                         's: {}'.format(logdet_I_A))

    result = _IntermediateSampleInfo(alpha_star=alpha_star,
                                     logdet_I_A=logdet_I_A,
                                     q=-1,
                                     s=s,
                                     z=z,
                                     rls_estimate=rls_estimate)

    return result


def vfx_sampling_do_sampling_loop(X_data, eval_L, intermediate_sample_info, rng, max_iter=1000, verbose=True, **kwargs):
    """Given pre-computed information, run a rejection sampling loop to generate DPP samples.
        :param array_like X_data: dataset such that L = eval_L(X_data), out of which we are sampling objects
        according to a DPP
        :param callable eval_L: likelihood function. Given two sets of n points X and m points Y, eval_L(X, Y) should
        compute the (n x m) matrix containing the likelihood between points. The function should also
        accept a single argument X and return eval_L(X) = eval_L(X, X).
        As an example, see the implementation of any of the kernels provided by scikit-learn
        (e.g. sklearn.gaussian_process.kernels.PairwiseKernel).
        :param _IntermediateSampleInfo: Pre-computed information necessary for the vfx rejection sampling loop,
        as returned by :func:`vfx_sampling_precompute_constants`
        :param RandomState rng: random source used for sampling
        :param max_iter:
        :type max_iter:
            int, default 1000
        :param bool verbose: controls verbosity of debug output, including progress bars.
        The progress bar reports the execution of the rejection sampling loop, showing:
            - acc_thresh: latest computed probability of acceptance
            - rej_iter: iteration of the rejection sampling loop (i.e. rejections so far)
        :type verbose:
            bool, default True
        :param dict kwargs: we add a unused catch all kwargs argument to make sure that the user can pass the
        same set of parameters to both vfx_sampling_precompute_constants and vfx_sampling_do_sampling_loop. This
        way if there is any spurious non-shared parameter (e.g. rls_oversample_bless) we simply ignore it.
        :return: Sample from a DPP (as a list) and number of rejections as int
        :rtype: tuple(list, int)
    """
    #TODO: taking as input a catch-all kwargs can be misleading for the user. e.g. if there is a typo in a paremater
    # it will silently ignore it and use the default instead

    n, d = X_data.shape

    # rename it to pre-computed state for shortness
    pc_state = intermediate_sample_info

    # Phase 3: rejection sampling loop

    with get_progress_bar(disable=not verbose) as prog_bar:
        for rej_iter in range(max_iter):
            # sample t
            lam = np.ceil(pc_state.q * np.exp(pc_state.s / pc_state.q))
            t = rng.poisson(lam=lam.astype('int'))

            # sample sigma subset
            sigma = rng.choice(n, size=t, p=pc_state.rls_estimate / pc_state.s, replace=True)
            sigma_uniq, sigma_uniq_count = np.unique(sigma, return_counts=True)
            X_sigma_uniq = X_data[sigma_uniq, :]

            # compute log(Det(I + \tilda{L}_sigma)) = log(Det(I + W*L_sigma*W))
            # with W_ii = ( s / (q * l_i) )^1/2
            # this is done by computing
            # log(Det(I + W*L_sigma*W))
            # = log(Det(W * (W^-2 + L_sigma) * W))
            # = log(Det(W) * Det(W^-2 + L_sigma) * Det(W))
            # = log(Det(W)) + log(Det(W^-2 + L_sigma)) + log(Det(W))
            # = log(Det(W^2)) + log(Det(W^-2 + L_sigma))
            # = -log(Det(W^-2)) + log(Det(W^-2 + L_sigma))

            W_square_inv = pc_state.q * pc_state.rls_estimate[sigma_uniq] / (pc_state.s * sigma_uniq_count)

            I_L_sigma = (pc_state.alpha_star * eval_L(X_sigma_uniq, X_sigma_uniq)
                         + np.diag(W_square_inv))

            s_logdet, logdet_I_L_sigma = np.linalg.slogdet(I_L_sigma)
            if not s_logdet >= 0.0:
                raise ValueError('logdet_I_L_sigma is negative, this should never happen. '
                                 's: {}'.format(s_logdet))

            logdet_W_square_inv = np.sum(np.log(W_square_inv))

            acc_thresh = (pc_state.z
                          + logdet_I_L_sigma
                          - logdet_W_square_inv
                          - pc_state.logdet_I_A
                          - t * pc_state.s / pc_state.q).item()

            accept = np.log(rng.rand()) <= acc_thresh

            prog_bar.set_postfix(acc_thresh=np.exp(acc_thresh), rej_count=rej_iter)
            prog_bar.update()

            if accept:
                break
        else:
            raise ValueError('The vfx sampler reached the maximum number of rejections allowed '
                             'for the intermediate sample selection ({}), try to increase the q factor '
                             '(see q_func parameter) or the Nystrom approximation accuracy '
                             'see rls_oversample_* parameters).'.format(max_iter))

    # Phase 4: use L_tilda to perform exact DPP sampling
    # compute alpha_star * L_tilda = alpha_star * W*L_sigma*W
    W = (np.sqrt(pc_state.s * sigma_uniq_count)
         / np.sqrt(pc_state.q * pc_state.rls_estimate[sigma_uniq]).reshape(-1, 1))

    L_tilda = pc_state.alpha_star * W.T * eval_L(X_sigma_uniq, X_sigma_uniq) * W

    E, U = np.linalg.eigh(L_tilda)

    # this has to be here rather than at the top to avoid circular dependencies
    # TODO: maybe refactor to avoid this
    from dppy.finite_dpps import FiniteDPP
    DPP = FiniteDPP(kernel_type='likelihood', L_eig_dec=(E, U))
    S_tilda = np.array(DPP.sample_exact(random_state=rng), dtype=int)

    S = sigma_uniq[S_tilda].ravel().tolist()

    return S, rej_iter



# coding: utf8
""" Implementation of the meta-class :class:`BetaEnsemble` see :math:`\\beta-`:ref:`Ensembles<beta_ensembles>` with children:

- :class:`HermiteEnsemble`
- :class:`LaguerreEnsemble`
- :class:`JacobiEnsemble`
- :class:`CircularEnsemble`
- :class:`GinibreEnsemble`

Such objects have 4 main methods:

- :py:func:`sample_full_model`
- :py:func:`sample_banded_model`
- :py:func:`plot` to display a scatter plot of the last sample and eventually the limiting distribution (after normalization)
- :py:func:`hist` to display a histogram of the last sample and eventually the limiting distribution (after normalization)

.. seealso:

    `Documentation on ReadTheDocs <https://dppy.readthedocs.io/en/latest/continuous_dpps/beta_ensembles.html>`_
"""

from abc import ABCMeta, abstractmethod

import numpy as np
from scipy import stats
import matplotlib.pyplot as plt

from re import findall  # to convert class names to string

import dppy.random_matrices as rm
from dppy.utils import check_random_state


class BetaEnsemble(metaclass=ABCMeta):
    """ :math:`\\beta`-Ensemble object parametrized by

    :param beta:
        :math:`\\beta >= 0` inverse temperature parameter.

        The default :py:attr:`beta`:math:`=2` corresponds to the DPP case,
        see :ref:`beta_ensembles_definition_OPE`
    :type beta:
        int, float, default :math:`2`

    .. seealso::

        - :math:`\\beta`-Ensembles :ref:`definition <beta_ensembles_definition>`
    """

    def __init__(self, beta=2):

        if not (beta >= 0):
            raise ValueError('`beta` must be >=0. Given: {}'.format(self.beta))
        self.beta = beta

        # Split object name at uppercase
        self.name = ' '.join(findall('[A-Z][^A-Z]*', self.__class__.__name__))
        self.params = {'size_N': 10}  # Number of points and ref measure params

        self.sampling_mode = ''
        self.list_of_samples = []

    @property
    def _str_title(self):
        return r'Realization of {} points of {} with $\beta={}$'.format(
            self.params['size_N'], self.name, self.beta)

    def __str__(self):
        str_info = ('{} with beta = {}'.format(self.name, self.beta),
                    'sampling parameters = {}'.format(self.params),
                    'number of samples = {}'.format(len(self.list_of_samples)))

        return '\n'.join(str_info)

    def flush_samples(self):
        """ Empty the :py:attr:`list_of_samples` attribute.
        """
        self.list_of_samples = []

    @abstractmethod
    def sample_full_model(self):
        """Sample from underlying :math:`\\beta`-Ensemble using the corresponding full matrix model.
        Arguments are the associated matrix dimensions
        """

    @abstractmethod
    def sample_banded_model(self):
        """Sample from underlying :math:`\\beta`-Ensemble using the corresponding banded matrix model.
        Arguments are the associated reference measure's parameters, or the matrix dimensions used in :py:meth:`sample_full_model`
        """

    @abstractmethod
    def plot(self):
        """Display last realization of the underlying :math:`\\beta`-Ensemble.
        For some :math:`\\beta`-Ensembles, a normalization argument is available to display the limiting (or equilibrium) distribution and scale the points accordingly.
        """

    @abstractmethod
    def hist(self):
        """Display histogram of the last realization of the underlying :math:`\\beta`-Ensemble.
        For some :math:`\\beta`-Ensembles, a normalization argument is available to display the limiting (or equilibrium) distribution and scale the points accordingly.
        """

    @abstractmethod
    def normalize_points(self):
        """ Normalize points ormalization argument is available to display the limiting (or equilibrium) distribution and scale the points accordingly.
        """


class HermiteEnsemble(BetaEnsemble):
    """ Hermite Ensemble object

    .. seealso::

        - :ref:`Full matrix model <hermite_full_matrix_model>` associated to the Hermite ensemble
        - :ref:`Tridiagonal matrix model <hermite_banded_matrix_model>` associated to the Hermite ensemble
    """

    def __init__(self, beta=2):

        super().__init__(beta=beta)

        # Default parameters for ``loc`` and ``scale`` correspond to the
        # reference measure N(0,2) in the full matrix model
        params = {'loc': 0.0, 'scale': np.sqrt(2.0),
                  'size_N': 10}
        self.params.update(params)

    def sample_full_model(self, size_N=10,
                          random_state=None):
        """ Sample from :ref:`full matrix model <Hermite_full_matrix_model>` associated to the Hermite ensemble.
        Only available for :py:attr:`beta` :math:`\\in\\{1, 2, 4\\}`
        and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the Gaussian :math:`\\mathcal{N}(\\mu,\\sigma^2)` reference measure

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized
        :type size_N:
            int, default :math:`10`

        .. note::

            The reference measure associated with the :ref:`full matrix model <hermite_full_matrix_model>` is :math:`\\mathcal{N}(0,2)`.
            For this reason, in the :py:attr:`sampling_params` attribute, the values of the parameters are set to ``loc``:math:`=0` and ``scale``:math:`=\\sqrt{2}`.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N`` parameter.

        .. seealso::

            - :ref:`Full matrix model <hermite_full_matrix_model>` associated to the Hermite ensemble
            - :py:meth:`sample_banded_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'full'
        params = {'loc': 0.0, 'scale': np.sqrt(2.0), 'size_N': size_N}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # Sample N i.i.d. gaussian N(0,2)
            sampl = rng.normal(loc=self.params['loc'],
                               scale=self.params['scale'],
                               size=self.params['size_N'])
        else:  # if beta > 0
            sampl = rm.hermite_sampler_full(N=self.params['size_N'],
                                            beta=self.beta,
                                            random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def sample_banded_model(self, loc=0.0, scale=np.sqrt(2.0), size_N=10,
                            random_state=None):
        """ Sample from :ref:`tridiagonal matrix model <hermite_banded_matrix_model>` associated to the Hermite Ensemble.
        Available for :py:attr:`beta` :math:`>0` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the Gaussian :math:`\\mathcal{N}(\\mu,\\sigma^2)` reference measure

        :param loc:
            Mean :math:`\\mu` of the Gaussian :math:`\\mathcal{N}(\\mu, \\sigma^2)`
        :type loc:
            float, default :math:`0`

        :param scale:
            Standard deviation :math:`\\sigma` of the Gaussian :math:`\\mathcal{N}(\\mu, \\sigma^2)`
        :type scale:
            float, default :math:`\\sqrt{2}`

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized
        :type size_N:
            int, default :math:`10`

        .. note::

            The reference measure associated with the :ref:`full matrix model <hermite_full_matrix_model>` is :math:`\\mathcal{N}(0,2)`.
            For this reason, in the :py:attr:`sampling_params` attribute, the default values are set to ``loc``:math:`=0` and ``scale``:math:`=\\sqrt{2}`.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N`` parameter.

        .. seealso::

            - :ref:`Tridiagonal matrix model <hermite_banded_matrix_model>` associated to the Hermite ensemble
            - :cite:`DuEd02` II-C
            - :py:meth:`sample_full_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'banded'
        params = {'loc': loc, 'scale': scale, 'size_N': size_N}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # Sample N i.i.d. gaussian N(mu, sigma^2)
            sampl = rng.normal(loc=self.params['loc'],
                               scale=self.params['scale'],
                               size=self.params['size_N'])
        else:  # if beta > 0
            sampl = rm.mu_ref_normal_sampler_tridiag(
                        loc=self.params['loc'],
                        scale=self.params['scale'],
                        beta=self.beta,
                        size=self.params['size_N'],
                        random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def normalize_points(self, points):
        """ Normalize points obtained after sampling to fit the limiting distribution, i.e., the semi-circle

        .. math::

            f(x) = \\frac{1}{2\\pi} \\sqrt{4-x^2}

        :param points:
            A sample from Hermite ensemble, accessible through the :py:attr:`list_of_samples` attribute
        :type points:
            array_like

        - If sampled using :py:meth:`sample_banded_model` with reference measure :math:`\\mathcal{N}(\\mu,\\sigma^2)`

            1. Normalize the points to fit the p.d.f. of :math:`\\mathcal{N}(0,2)` reference measure of the :ref:`full matrix model <hermite_full_matrix_model>`

                .. math::

                    x \\mapsto \\sqrt{2}\\frac{x-\\mu}{\\sigma}

            2. If :py:attr:`beta` :math:`>0`, normalize the points to fit the semi-circle distribution

                .. math::

                    x \\mapsto \\frac{x}{\\beta N}

                Otherwise if :py:attr:`beta` :math:`=0` do nothing more

        - If sampled using :py:meth:`sample_full_model`, apply 2. above

        .. note::

            This method is called in :py:meth:`plot` and :py:meth:`hist` when ``normalization=True``
        """

        if self.sampling_mode == 'banded':
            # Normalize to fit N(0,2) ref measure of the full matrix model
            points -= self.params['loc']
            points /= np.sqrt(0.5) * self.params['scale']
        else:  # 'full'
            pass

        if self.beta > 0:
            # Normalize to fit the semi-circle distribution
            points /= np.sqrt(self.beta * self.params['size_N'])
        else:
            pass

        return points

    def __display_and_normalization(self, display_type, normalization):

        if not self.list_of_samples:
            raise ValueError('Empty `list_of_samples`,sample first!')
        else:
            points = self.list_of_samples[-1].copy()  # Pick last sample
            if normalization:
                points = self.normalize_points(points)

        fig, ax = plt.subplots(1, 1)
        # Title
        str_beta = '' if self.beta > 0 else 'with i.i.d. draws'
        title = '\n'.join([self._str_title, str_beta])
        plt.title(title)

        if self.beta == 0:
            if normalization:
                # Display N(0,2) reference measure of the full matrix model
                mu, sigma = 0.0, np.sqrt(2.0)
                x = mu + 3.5 * sigma * np.linspace(-1, 1, 100)
                ax.plot(x, stats.norm.pdf(x, mu, sigma),
                        'r-', lw=2, alpha=0.6,
                        label=r'$\mathcal{{N}}(0,2)$')
            else:
                pass

        else:  # self.beta > 0
            if normalization:
                # Display the limiting distribution: semi-circle law
                x = np.linspace(-2, 2, 100)
                ax.plot(x, rm.semi_circle_law(x),
                        'r-', lw=2, alpha=0.6,
                        label=r'$f_{semi-circle}$')
            else:
                pass

        if display_type == 'scatter':
            ax.scatter(points, np.zeros_like(points),
                       c='blue',
                       label='sample')

        elif display_type == 'hist':
            ax.hist(points, bins=30, density=1,
                    facecolor='blue', alpha=0.5,
                    label='hist')
        else:
            pass

        plt.legend(loc='best', frameon=False)

    def plot(self, normalization=True):
        """ Display the last realization of the :class:`HermiteEnsemble` object

        :param normalization:
            When ``True``, the points are first normalized (see :py:meth:`normalize_points`) so that they concentrate as

            - If :py:attr:`beta` :math:`=0`, the :math:`\\mathcal{N}(0, 2)` reference measure associated to full :ref:`full matrix model <hermite_full_matrix_model>`
            - If :py:attr:`beta` :math:`>0`, the limiting distribution, i.e., the semi-circle distribution

            in both cases, the corresponding p.d.f. is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`normalize_points`
            - :py:meth:`hist`
            - :ref:`Full matrix model <hermite_full_matrix_model>` associated to the Hermite ensemble
            - :ref:`Tridiagonal matrix model <hermite_banded_matrix_model>` associated to the Hermite ensemble
        """

        self.__display_and_normalization('scatter', normalization)

    def hist(self, normalization=True):
        """ Display the histogram of the last realization of the :class:`HermiteEnsemble` object.

        :param normalization:
            When ``True``, the points are first normalized (see :py:meth:`normalize_points`) so that they concentrate as

            - If :py:attr:`beta` :math:`=0`, the :math:`\\mathcal{N}(0, 2)` reference measure associated to full :ref:`full matrix model <hermite_full_matrix_model>`
            - If :py:attr:`beta` :math:`>0`, the limiting distribution, i.e., the semi-circle distribution

            in both cases, the corresponding p.d.f. is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`normalize_points`
            - :py:meth:`plot`
            - :ref:`Full matrix model <hermite_full_matrix_model>` associated to the Hermite ensemble
            - :ref:`Tridiagonal matrix model <hermite_banded_matrix_model>` associated to the Hermite ensemble
        """
        self.__display_and_normalization('hist', normalization)


class LaguerreEnsemble(BetaEnsemble):
    """ Laguerre Ensemble object

    .. seealso::

        - :ref:`Full matrix model <laguerre_full_matrix_model>` associated to the Laguerre ensemble
        - :ref:`Tridiagonal matrix model <laguerre_banded_matrix_model>` associated to the Laguerre ensemble
    """

    def __init__(self, beta=2):

        super().__init__(beta=beta)

        params = {'shape': 0.0, 'scale': 2.0,
                  'size_N': 10, 'size_M': None}
        self.params.update(params)

    def sample_full_model(self, size_N=10, size_M=100,
                          random_state=None):
        """ Sample from :ref:`full matrix model <Laguerre_full_matrix_model>` associated to the Laguerre ensemble. Only available for :py:attr:`beta` :math:`\\in\\{1, 2, 4\\}` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the :math:`\\Gamma(k,\\theta)` reference measure

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized.
            First dimension of the matrix used to form the covariance matrix to be diagonalized, see :ref:`full matrix model <laguerre_full_matrix_model>`.
        :type size_N:
            int, default :math:`10`

        :param size_M:
            Second dimension :math:`M` of the matrix used to form the covariance matrix to be diagonalized, see :ref:`full matrix model <laguerre_full_matrix_model>`.

        :type size_M:
            int, default :math:`100`

        .. note::

            The reference measure associated with the :ref:`full matrix model <laguerre_full_matrix_model>` is :math:`\\Gamma\\left(\\frac{\\beta}{2}(M-N+1), 2\\right)`.
            For this reason, in the :py:attr:`sampling_params`, the values of the parameters are set to ``shape``:math:`=\\frac{\\beta}{2}(M-N+1)` and ``scale``:math:`=2`.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N`` and ``size_M`` parameters.

        .. seealso::

            - :ref:`Full matrix model <Laguerre_full_matrix_model>` associated to the Laguerre ensemble
            - :py:meth:`sample_banded_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'full'

        if size_M >= size_N:
            # Define the parameters of the associated Gamma distribution
            shape, scale = 0.5 * self.beta * (size_M - size_N + 1), 2.0
        else:
            err_print = ('Must have M >= N.',
                         'Given: M={} < N={}'.format(size_M, size_N))
            raise ValueError(' '.join(err_print))

        params = {'shape': shape, 'scale': scale,
                  'size_N': size_N, 'size_M': size_M}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # rng.gamma(shape=0,...) when doesn't return error! contrary to sp.stats.gamma(a=0).rvs(), see https://github.com/numpy/numpy/issues/12367
            # sampl = stats.gamma.rvs(a=params['shape'], loc=0.0, scale=params['scale'],                                       size=params['size_N'])
            if params['shape'] > 0:
                sampl = rng.gamma(shape=self.params['shape'],
                                  scale=self.params['scale'],
                                  size=self.params['size_N'])
            else:
                err_print = ('shape<=0.',
                             'Here beta=0, hence shape=beta/2*(M-N+1)=0')
                raise ValueError(' '.join(err_print))

        else:  # if beta > 0
            sampl = rm.laguerre_sampler_full(M=self.params['size_M'],
                                             N=self.params['size_N'],
                                             beta=self.beta,
                                             random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def sample_banded_model(self,
                            shape=1.0, scale=2.0,
                            size_N=10, size_M=None,
                            random_state=None):
        """ Sample from :ref:`tridiagonal matrix model <Laguerre_banded_matrix_model>` associated to the Laguerre ensemble. Available for :py:attr:`beta` :math:`>0` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the :math:`\\Gamma(k,\\theta)` reference measure

        :param shape:
            Shape parameter :math:`k` of :math:`\\Gamma(k, \\theta)` reference measure
        :type shape:
            float, default :math:`1`

        :param scale:
            Scale parameter :math:`\\theta` of :math:`\\Gamma(k, \\theta)` reference measure
        :type scale:
            float, default :math:`2.0`

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized.
            Equivalent to the first dimension :math:`N` of the matrix used to form the covariance matrix in the :ref:`full matrix model <laguerre_full_matrix_model>`.
        :type size_N:
            int, default :math:`10`

        :param size_M:
            Equivalent to the second dimension :math:`M` of the matrix used to form the covariance matrix in the :ref:`full matrix model <laguerre_full_matrix_model>`.

        :type size_M:
            int, default None

        - If ``size_M`` is not provided:

            In the :py:attr:`sampling_params`, ``size_M`` is set to
            ``size_M``:math:`= \\frac{2k}{\\beta} + N - 1`, to give an idea of the corresponding second dimension :math:`M`.


        - If ``size_M`` is provided:

            In the :py:attr:`sampling_params`, ``shape`` and ``scale`` are set to:
            ``shape``:math:`=\\frac{1}{2} \\beta (M-N+1)` and ``scale``:math:`=2`


        .. note::

            The reference measure associated with the :ref:`full matrix model <laguerre_full_matrix_model>` is :math:`\\Gamma\\left(\\frac{\\beta}{2}(M-N+1), 2\\right)`. This explains the role of the ``size_M`` parameter.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N`` and ``size_M`` parameters

        .. seealso::

            - :ref:`Tridiagonal matrix model <Laguerre_banded_matrix_model>` associated to the Laguerre ensemble
            - :cite:`DuEd02` III-B
            - :py:meth:`sample_full_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'banded'

        if not size_M:  # Default setting
            if self.beta > 0:
                size_M = 2 / self.beta * shape + size_N - 1
            else:
                size_M = np.inf

        elif size_M >= size_N:
            # define the parameters of the associated Gamma distribution
            shape, scale = 0.5 * self.beta * (size_M - size_N + 1), 2.0

        else:
            err_print = ('Must have M >= N.',
                         'Given: M={} < N={}'.format(size_M, size_N))
            raise ValueError(' '.join(err_print))

        params = {'shape': shape, 'scale': scale,
                  'size_N': size_N, 'size_M': size_M}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # rng.gamma(shape=0,...) when doesn't return error! contrary to sp.stats.gamma(a=0).rvs(), see https://github.com/numpy/numpy/issues/12367
            # sampl = stats.gamma.rvs(a=params['shape'], loc=0.0, scale=params['scale'], size=params['size_N'])
            if params['shape'] > 0:
                sampl = rng.gamma(shape=self.params['shape'],
                                  scale=self.params['scale'],
                                  size=self.params['size_N'])
            else:
                err_print = ('shape<=0.',
                             'Here beta=0, hence shape=beta/2*(M-N+1)=0')
                raise ValueError(' '.join(err_print))

        else:  # if beta > 0
            sampl = rm.mu_ref_gamma_sampler_tridiag(shape=self.params['shape'],
                                                    scale=self.params['scale'],
                                                    beta=self.beta,
                                                    size=self.params['size_N'],
                                                    random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def normalize_points(self, points):
        """ Normalize points obtained after sampling to fit the limiting distribution, i.e., the Marcenko-Pastur distribution

        .. math::

            \\frac{1}{2\\pi}
            \\frac{\\sqrt{(\\lambda_+-x)(x-\\lambda_-)}}{cx}
            1_{[\\lambda_-,\\lambda_+]}
            dx

        where :math:`c = \\frac{M}{N}` and :math:`\\lambda_\\pm = (1\\pm\\sqrt{c})^2`

        :param points:
            A sample from Laguerre ensemble, accessible through the :py:attr:`list_of_samples` attribute
        :type points:
            array_like

        - If sampled using :py:meth:`sample_banded_model` with reference measure :math:`\\Gamma(k,\\theta)`

            .. math::

                x \\mapsto \\frac{2x}{\\theta}
                \\quad \\text{and} \\quad
                x \\mapsto \\frac{x}{\\beta M}

        - If sampled using :py:meth:`sample_full_model`

            .. math::

                x \\mapsto \\frac{x}{\\beta M}

        .. note::

            This method is called in :py:meth:`plot` and :py:meth:`hist` when ``normalization=True``.
        """

        if self.sampling_mode == 'banded':
            # Normalize to fit Gamma(*,2) refe measure of the full matrix model
            points /= 0.5 * self.params['scale']
        else:  # self.sampling_mode == 'full':
            pass

        if self.beta > 0:
            # Normalize to fit the Marcenko-Pastur distribution
            points /= self.beta * self.params['size_M']
        else:
            pass
            # set a warning, won't concentrate as semi-circle, they are i.i.d.

        return points

    def __display_and_normalization(self, display_type, normalization):

        if not self.list_of_samples:
            raise ValueError('Empty `list_of_samples`, sample first!')
        else:
            points = self.list_of_samples[-1].copy()  # Pick last sample
            if normalization:
                points = self.normalize_points(points)

        N, M = [self.params[key] for key in ['size_N', 'size_M']]

        fig, ax = plt.subplots(1, 1)
        # Title
        str_ratio = r'with ratio $M/N \approx {:.3f}$'.format(M / N)
        # Answers Issue #33 raised by @adrienhardy
        str_beta = '' if self.beta > 0 else 'with i.i.d. draws'
        title = '\n'.join([self._str_title,
                           ' '.join([str_ratio, str_beta])])
        plt.title(title)

        if self.beta == 0:
            if normalization:
                # Display Gamma(k,2) reference measure of the full matrix model
                k, theta = self.params['shape'], 2
                x = np.linspace(0, np.max(points) + 4, 100)
                ax.plot(x, stats.gamma.pdf(x, a=k, loc=0.0, scale=theta),
                        'r-', lw=2, alpha=0.6,
                        label=r'$\Gamma({},{})$'.format(k, theta))

        else:  # self.beta > 0
            if normalization:
                # Display the limiting distribution: Marcenko-Pastur
                x = np.linspace(1e-2, np.max(points) + 0.3, 100)
                ax.plot(x, rm.marcenko_pastur_law(x, M, N),
                        'r-', lw=2, alpha=0.6,
                        label=r'$f_{Marcenko-Pastur}$')

        if display_type == 'scatter':
            ax.scatter(points, np.zeros_like(points),
                       c='blue',
                       label='sample')

        elif display_type == 'hist':
            ax.hist(points, bins=30, density=1,
                    facecolor='blue', alpha=0.5,
                    label='hist')
        else:
            pass

        plt.legend(loc='best', frameon=False)

    def plot(self, normalization=True):
        """ Display the last realization of the :class:`LaguerreEnsemble` object

        :param normalization:
            When ``True``, the points are first normalized (see :py:meth:`normalize_points`) so that they concentrate as

            - If :py:attr:`beta` :math:`=0`, the :math:`\\Gamma(k, 2)` reference measure associated to full :ref:`full matrix model <laguerre_full_matrix_model>`
            - If :py:attr:`beta` :math:`>0`, the limiting distribution, i.e., the Marcenko-Pastur distribution

            in both cases the corresponding p.d.f. is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`normalize_points`
            - :py:meth:`hist`
            - :ref:`Full matrix model <Laguerre_full_matrix_model>` associated to the Laguerre ensemble
            - :ref:`Tridiagonal matrix model <Laguerre_banded_matrix_model>` associated to the Laguerre ensemble
        """

        self.__display_and_normalization('scatter', normalization)

    def hist(self, normalization=True):
        """ Display the histogram of the last realization of the :class:`LaguerreEnsemble` object.

        :param normalization:
            When ``True``, the points are first normalized (see :py:meth:`normalize_points`) so that they concentrate as

            - If :py:attr:`beta` :math:`=0`, the :math:`\\Gamma(k, 2)` reference measure associated to full :ref:`full matrix model <laguerre_full_matrix_model>`
            - If :py:attr:`beta` :math:`>0`, the limiting distribution, i.e., the Marcenko-Pastur distribution

            in both cases the corresponding p.d.f. is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`normalize_points`
            - :py:meth:`plot`
            - :ref:`Full matrix model <Laguerre_full_matrix_model>` associated to the Laguerre ensemble
            - :ref:`Tridiagonal matrix model <Laguerre_banded_matrix_model>` associated to the Laguerre ensemble
        """
        self.__display_and_normalization('hist', normalization)


class JacobiEnsemble(BetaEnsemble):
    """ Jacobi Ensemble object

    .. seealso::

        - :ref:`Full matrix model <jacobi_full_matrix_model>` associated to the Jacobi ensemble
        - :ref:`Tridiagonal matrix model <jacobi_banded_matrix_model>` associated to the Jacobi ensemble
    """

    def __init__(self, beta=2):

        super().__init__(beta=beta)

        params = {'a': 1.0, 'b': 1.0,
                  'size_N': 10, 'size_M1': None, 'size_M2': None}
        self.params.update(params)

    def sample_full_model(self, size_N=100, size_M1=150, size_M2=200,
                          random_state=None):
        """ Sample from :ref:`full matrix model <Jacobi_full_matrix_model>` associated to the Jacobi ensemble. Only available for :py:attr:`beta` :math:`\\in\\{1, 2, 4\\}` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the :math:`\\operatorname{Beta}(a,b)` reference measure

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized.
            First dimension of the matrix used to form the covariance matrix to be diagonalized, see :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_N:
            int, default :math:`100`

        :param size_M1:
            Second dimension :math:`M_1` of the first matrix used to form the matrix to be diagonalized, see :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_M1:
            int, default :math:`150`

        :param size_M2:
            Second dimension :math:`M_2` of the second matrix used to form the matrix to be diagonalized, see :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_M2:
            int, default :math:`200`

        .. note::

            The reference measure associated with the :ref:`full matrix model <jacobi_full_matrix_model>` is

            .. math::

                \\operatorname{Beta}\\left(\\frac{\\beta}{2}(M_1-N+1), \\frac{\\beta}{2}(M_2-N+1)\\right)

            For this reason, in the :py:attr:`sampling_params` attribute, the values of the parameters are set to ``a``:math:`=\\frac{\\beta}{2}(M_1-N+1)` and ``b``:math:`=\\frac{\\beta}{2}(M_2-N+1)`.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N``, ``size_M2`` and ``size_M2`` parameters.

        .. seealso::

            - :ref:`Full matrix model <Jacobi_full_matrix_model>` associated to the Jacobi ensemble
            - :py:meth:`sample_banded_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'full'

        if (size_M1 >= size_N) and (size_M2 >= size_N):
            # all([var >= size_N for var in [size_M1, size_M2]]
            a = 0.5 * self.beta * (size_M1 - size_N + 1)
            b = 0.5 * self.beta * (size_M2 - size_N + 1)

        else:
            err_print = ('Must have M1, M2 >= N.',
                         'Given: M1={}, M2={} and N={}'.format(size_M1,
                                                               size_M2,
                                                               size_N))
            raise ValueError(' '.join(err_print))

        params = {'a': a, 'b': b,
                  'size_N': size_N, 'size_M1': size_M1, 'size_M2': size_M2}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # Sample i.i.d. Beta(a,b) if size_M1,2 were used a,b = beta/2 (M_1,2 - N + 1) = 0 => ERROR
            sampl = rng.beta(a=self.params['a'],
                             b=self.params['b'],
                             size=self.params['size_N'])
        else:
            sampl = rm.jacobi_sampler_full(M_1=self.params['size_M1'],
                                           M_2=self.params['size_M2'],
                                           N=self.params['size_N'],
                                           beta=self.beta,
                                           random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def sample_banded_model(self, a=1.0, b=2.0,
                            size_N=10, size_M1=None, size_M2=None,
                            random_state=None):
        """ Sample from :ref:`tridiagonal matrix model <Jacobi_banded_matrix_model>` associated to the Jacobi ensemble. Available for :py:attr:`beta` :math:`>0` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. points from the :math:`\\operatorname{Beta}(a,b)` reference measure

        :param shape:
            Shape parameter :math:`k` of :math:`\\Gamma(k, \\theta)` reference measure
        :type shape:
            float, default :math:`1`

        :param scale:
            Scale parameter :math:`\\theta` of :math:`\\Gamma(k, \\theta)` reference measure
        :type scale:
            float, default :math:`2.0`

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized.
            Equivalent to the first dimension :math:`N` of the matrices used in the :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_N:
            int, default :math:`10`

        :param size_M1:
            Equivalent to the second dimension :math:`M_1` of the first matrix used in the :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_M1:
            int

        :param size_M2:
            Equivalent to the second dimension :math:`M_2` of the second matrix used in the :ref:`full matrix model <jacobi_full_matrix_model>`.
        :type size_M2:
            int

        .. note::

            The reference measure associated with the :ref:`full matrix model <jacobi_full_matrix_model>` is :

            .. math::

                \\operatorname{Beta}\\left(\\frac{\\beta}{2}(M_1-N+1), \\frac{\\beta}{2}(M_2-N+1)\\right)

            For this reason, in the :py:attr:`sampling_params` attribute, the values of the parameters are set to ``a``:math:`=\\frac{\\beta}{2}(M_1-N+1)` and ``b``:math:`=\\frac{\\beta}{2}(M_2-N+1)`.

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N``, ``size_M2`` and ``size_M2`` parameters.

        - If ``size_M1`` and ``size_M2`` are not provided:

            In the :py:attr:`sampling_params` attribute, ``size_M1,2`` are set to
            ``size_M1``:math:`= \\frac{2a}{\\beta} + N - 1` and ``size_M2``:math:`= \\frac{2b}{\\beta} + N - 1`, to give an idea of the corresponding second dimensions :math:`M_{1,2}`.

        - If ``size_M1`` and ``size_M2`` are provided:

            In the :py:attr:`sampling_params` attribute, ``a`` and ``b`` are set to:
            ``a``:math:`=\\frac{\\beta}{2}(M_1-N+1)` and
            ``b``:math:`=\\frac{\\beta}{2}(M_2-N+1)`.

        .. seealso::

            - :ref:`Tridiagonal matrix model <Jacobi_banded_matrix_model>` associated to the Jacobi ensemble
            - :cite:`KiNe04` Theorem 2
            - :py:meth:`sample_full_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'banded'

        if not (size_M1 and size_M2):  # default setting

            if self.beta > 0:
                size_M1 = 2 / self.beta * a + size_N - 1
                size_M2 = 2 / self.beta * b + size_N - 1

            else:
                size_M1, size_M2 = np.inf, np.inf

        elif (size_M1 >= size_N) and (size_M2 >= size_N):
            # all([var >= size_N for var in [size_M1, size_M2]]
            a = 0.5 * self.beta * (size_M1 - size_N + 1)
            b = 0.5 * self.beta * (size_M2 - size_N + 1)

        else:
            err_print = ('Must have M1, M2 >= N.',
                         'Given: M1={}, M2={} and N={}'.format(size_M1,
                                                               size_M2,
                                                               size_N))
            raise ValueError(' '.join(err_print))

        params = {'a': a, 'b': b,
                  'size_N': size_N, 'size_M1': size_M1, 'size_M2': size_M2}
        self.params.update(params)

        if self.beta == 0:  # Answer issue #28 raised by @rbardenet
            # Sample i.i.d. Beta(a,b)
            # If size_M1,2 is used a, b = beta/2 (M1,2 - N + 1) = 0 => ERROR
            sampl = rng.beta(a=self.params['a'],
                             b=self.params['b'],
                             size=self.params['size_N'])
        else:
            sampl = rm.mu_ref_beta_sampler_tridiag(a=self.params['a'],
                                                   b=self.params['b'],
                                                   beta=self.beta,
                                                   size=self.params['size_N'],
                                                   random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def normalize_points(self, points):
        """No need to renormalize the points
        """
        return points

    def __display_and_normalization(self, display_type, normalization):

        if not self.list_of_samples:
            raise ValueError('Empty `list_of_samples`, sample first!')
        else:
            points = self.list_of_samples[-1].copy()  # Pick last sample

        N, M_1, M_2 = [self.params[key]
                       for key in ['size_N', 'size_M1', 'size_M2']]

        fig, ax = plt.subplots(1, 1)
        # Title, answers Issue #33 raised by @adrienhardy
        str_ratio = ', '.join(['with ratios',
                               r'$M_1/N \approx {:.3f}$'.format(M_1 / N),
                               r'$M_2/N \approx {:.3f}$'.format(M_2 / N)])
        str_beta = '' if self.beta > 0 else 'with i.i.d. draws'
        title = '\n'.join([self._str_title, ' '.join([str_ratio, str_beta])])
        plt.title(title)

        if self.beta == 0:
            if normalization:
                # Display Beta(a,b) reference measure
                a, b = [self.params[key] for key in ['a', 'b']]
                x = np.linspace(0, 1, 100)
                ax.plot(x, stats.beta.pdf(x, a=a, b=b),
                        'r-', lw=2, alpha=0.6,
                        label=r'$\operatorname{{Beta}}({},{})$'.format(a, b))
        else:  # self.beta > 0
            if normalization:
                # Display the limiting distribution: Wachter law
                eps = 5e-3
                x = np.linspace(eps, 1.0 - eps, 500)
                ax.plot(x, rm.wachter_law(x, M_1, M_2, N),
                        'r-', lw=2, alpha=0.6,
                        label=r'$f_{Wachter}$')

        if display_type == 'scatter':
            ax.scatter(points, np.zeros_like(points),
                       c='blue',
                       label='sample')

        elif display_type == 'hist':
            ax.hist(points, bins=30, density=1,
                    facecolor='blue', alpha=0.5,
                    label='hist')
        else:
            pass

        plt.legend(loc='best', frameon=False)

    def plot(self, normalization=True):
        """ Display the last realization of the :class:`JacobiEnsemble` object

        :param normalization:
            When ``True``

            - If :py:attr:`beta` :math:`=0`, display the p.d.f. of the :math:`\\operatorname{Beta}(a, b)`
            - If :py:attr:`beta` :math:`>0`, display the limiting distribution, i.e., the Wachter distribution

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`hist`
            - :ref:`Full matrix model <Jacobi_full_matrix_model>` associated to the Jacobi ensemble
            - :ref:`Tridiagonal matrix model <Jacobi_banded_matrix_model>` associated to the Jacobi ensemble
        """

        self.__display_and_normalization('scatter', normalization)

    def hist(self, normalization=True):
        """ Display the histogram of the last realization of the :class:`JacobiEnsemble` object.

        :param normalization:
            When ``True``

            - If :py:attr:`beta` :math:`=0`, display the p.d.f. of the :math:`\\operatorname{Beta}(a, b)`
            - If :py:attr:`beta` :math:`>0`, display the limiting distribution, i.e., the Wachter distribution

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`normalize_points`
            - :py:meth:`plot`
            - :ref:`Full matrix model <Jacobi_full_matrix_model>` associated to the Jacobi ensemble
            - :ref:`Tridiagonal matrix model <Jacobi_banded_matrix_model>` associated to the Jacobi ensemble
        """
        self.__display_and_normalization('hist', normalization)


class CircularEnsemble(BetaEnsemble):
    """ Circular Ensemble object

    .. seealso::

        - :ref:`Full matrix model <circular_full_matrix_model>` associated to the Circular ensemble
        - :ref:`Quindiagonal matrix model <circular_banded_matrix_model>` associated to the Circular ensemble
    """

    def __init__(self, beta=2):

        if not isinstance(beta, int):
            raise ValueError('`beta` must be int >0. Given: {}'.format(beta))
        super().__init__(beta=beta)

        params = {'size_N': 10}
        self.params.update(params)

    def sample_full_model(self, size_N=10, haar_mode='Hermite',
                          random_state=None):
        """ Sample from :ref:`tridiagonal matrix model <Circular_full_matrix_model>` associated to the Circular ensemble. Only available for :py:attr:`beta` :math:`\\in\\{1, 2, 4\\}` and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. uniform points on the unit circle

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized
        :type size_N:
            int, default :math:`10`

        :param haar_mode:
            Sample Haar measure i.e. uniformly on the orthogonal/unitary/symplectic group using:
            - 'QR',
            - 'Hermite'
        :type haar_mode:
            str, default 'hermite'

        .. seealso::

            - :ref:`Full matrix model <circular_full_matrix_model>` associated to the Circular ensemble
            - :py:meth:`sample_banded_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'full'
        params = {'size_N': size_N, 'haar_mode': haar_mode}
        self.params.update(params)

        if self.beta == 0:  # i.i.d. points uniformly on the circle
            # Answer issue #28 raised by @rbardenet
            sampl = np.exp(2 * 1j * np.pi * rng.rand(params['size_N']))
        else:
            sampl = rm.circular_sampler_full(
                        N=self.params['size_N'],
                        beta=self.beta,
                        haar_mode=self.params['haar_mode'],
                        random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def sample_banded_model(self, size_N=10,
                            random_state=None):
        """ Sample from :ref:`Quindiagonal matrix model <Circular_banded_matrix_model>` associated to the Circular Ensemble.
        Available for :py:attr:`beta` :math:`\\in\\mathbb{N}^*`, and the degenerate case :py:attr:`beta` :math:`=0` corresponding to i.i.d. uniform points on the unit circle

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized
        :type size_N:
            int, default :math:`10`

        .. note::

            To compare :py:meth:`sample_banded_model` with :py:meth:`sample_full_model` simply use the ``size_N`` parameter.

        .. seealso::

            - :ref:`Quindiagonal matrix model <circular_banded_matrix_model>` associated to the Circular ensemble
            - :py:meth:`sample_full_model`
        """
        rng = check_random_state(random_state)

        self.sampling_mode = 'banded'
        params = {'size_N': size_N}
        self.params.update(params)

        if self.beta == 0:  # i.i.d. points uniformly on the circle
            # Answer issue #28 raised by @rbardenet
            sampl = np.exp(2 * 1j * np.pi * rng.rand(params['size_N']))
        else:
            sampl = rm.mu_ref_unif_unit_circle_sampler_quindiag(
                        beta=self.beta,
                        size=self.params['size_N'],
                        random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def normalize_points(self, points):
        """No need to renormalize the points
        """
        return points

    def __display_and_normalization(self, display_type, normalization):

        if not self.list_of_samples:
            raise ValueError('Empty `list_of_samples`, sample first!')
        else:
            points = self.list_of_samples[-1].copy()  # Pick last sample

        fig, ax = plt.subplots(1, 1)

        # Title
        samp_mod = ''

        if self.beta == 0:
            samp_mod = r'i.i.d samples from $\mathcal{U}_{[0,2\pi]}$'
        elif self.sampling_mode == 'full':
            samp_mod = 'full matrix model with haar_mode={}'.format(
                self.params['haar_mode'])
        else:  # self.sampling_mode == 'banded':
            samp_mod = 'quindiag model'

        title = '\n'.join([self._str_title, 'using {}'.format(samp_mod)])
        plt.title(title)

        if display_type == 'scatter':

            if normalization:
                # Draw unit circle
                unit_circle = plt.Circle((0, 0), 1, color='r', fill=False)
                ax.add_artist(unit_circle)

            ax.set_xlim([-1.3, 1.3])
            ax.set_ylim([-1.3, 1.3])
            ax.set_aspect('equal')
            ax.scatter(points.real, points.imag, c='blue', label='sample')

        elif display_type == 'hist':
            points = np.angle(points)

            if normalization:
                # Uniform distribution on [0, 2pi]
                ax.axhline(y=1 / (2 * np.pi),
                           color='r',
                           label=r'$\frac{1}{2\pi}$')

            ax.hist(points, bins=30, density=1,
                    facecolor='blue', alpha=0.5,
                    label='hist')
        else:
            pass

        plt.legend(loc='best', frameon=False)

    def plot(self, normalization=True):
        """ Display the last realization of the :class:`CircularEnsemble` object.

        :param normalization:
            When ``True``, the unit circle is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`hist`
            - :ref:`Full matrix model <circular_full_matrix_model>` associated to the Circular ensemble
            - :ref:`Quindiagonal matrix model <circular_banded_matrix_model>` associated to the Circular ensemble
        """

        self.__display_and_normalization('scatter', normalization)

    def hist(self, normalization=True):
        """ Display the histogram of the angles :math:`\\theta_{1}, \\dots, \\theta_{N}` associated to the last realization :math:`\\left\\{ e^{i \\theta_{1}}, \\dots, e^{i \\theta_{N}} \\right\\}`of the :class:`CircularEnsemble` object.

        :param normalization:
            When ``True``, the limiting distribution of the angles, i.e., the uniform distribution in :math:`[0, 2\\pi]` is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`sample_full_model`, :py:meth:`sample_banded_model`
            - :py:meth:`plot`
            - :ref:`Full matrix model <circular_full_matrix_model>` associated to the Circular ensemble
            - :ref:`Quindiagonal matrix model <circular_banded_matrix_model>` associated to the Circular ensemble
        """
        self.__display_and_normalization('hist', normalization)


class GinibreEnsemble(BetaEnsemble):
    """ Ginibre Ensemble object

    .. seealso::

        - :ref:`Full matrix model <ginibre_full_matrix_model>` associated to the Ginibre ensemble
    """

    def __init__(self, beta=2):

        if beta != 2:
            err_print = ('Ginibre ensemble is only available for `beta`=2.',
                         'Given {}'.format(beta))
            raise ValueError(' '.join(err_print))
        super().__init__(beta=beta)

        params = {'size_N': 10}
        self.params.update(params)

    def sample_full_model(self, size_N=10,
                          random_state=None):
        """ Sample from :ref:`full matrix model <Ginibre_full_matrix_model>` associated to the Ginibre ensemble. Only available for :py:attr:`beta` :math:`=2`

        :param size_N:
            Number :math:`N` of points, i.e., size of the matrix to be diagonalized
        :type size_N:
            int, default :math:`10`

        .. seealso::

            - :ref:`Full matrix model <ginibre_full_matrix_model>` associated to the Ginibre ensemble
        """
        rng = check_random_state(random_state)

        self.params.update({'size_N': size_N})
        sampl = rm.ginibre_sampler_full(N=self.params['size_N'],
                                        random_state=rng)

        self.list_of_samples.append(sampl)
        return sampl

    def sample_banded_model(self, *args, **kwargs):
        """ No banded model is known for Ginibre, use :py:meth:`sample_full_model`
        """
        raise NotImplementedError('No banded model is known for Ginibre, use `sample_full_model`')

    def normalize_points(self, points):
        """ Normalize points to concentrate in the unit disk.

            .. math::

                x \\mapsto \\frac{x}{\\sqrt{N}}

        .. seealso::

            - :py:meth:`plot`
            - :py:meth:`hist`
        """

        return points / np.sqrt(self.params['size_N'])

    def __display_and_normalization(self, display_type, normalization):

        if not self.list_of_samples:
            raise ValueError('Empty `list_of_samples`, sample first!')
        else:
            points = self.list_of_samples[-1].copy()  # Pick last sample
            if normalization:
                points = self.normalize_points(points)

        fig, ax = plt.subplots(1, 1)

        title = self._str_title

        if display_type == 'scatter':

            if normalization:
                # Draw unit circle
                unit_circle = plt.Circle((0, 0), 1, color='r', fill=False)
                ax.add_artist(unit_circle)

                ax.set_xlim([-1.3, 1.3])
                ax.set_ylim([-1.3, 1.3])
                ax.set_aspect('equal')

            ax.scatter(points.real, points.imag, c='blue', label='sample')

        elif display_type == 'hist':
            title = '\n'.join([title,
                               'Histogram of the modulus of each points'])

            if normalization:
                ax.plot([0, 1, 1], [0, 2, 0], color='r')

            bins = np.linspace(0, 1, 20)
            ax.hist(np.abs(points), bins=bins, density=1,
                    facecolor='blue', alpha=0.5,
                    label='hist')
        else:
            pass

        plt.title(title)

        # plt.legend(loc='best', frameon=False)

    def plot(self, normalization=True):
        """ Display the last realization of the :class:`GinibreEnsemble` object

        :param normalization:
            When ``True``, the points are first normalized so as to concentrate in the unit disk (see :py:meth:`normalize_points`) and the unit circle is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`normalize_points`
            - :py:meth:`sample_full_model`
            - :ref:`Full matrix model <ginibre_full_matrix_model>` associated to the Ginibre ensemble ensemble
        """

        self.__display_and_normalization('scatter', normalization)

    def hist(self, normalization=True):
        """ Display the histogram of the radius of the points the last realization of the :class:`GinibreEnsemble` object

        :param normalization:
            When ``True``, the points are first normalized so as to concentrate in the unit disk (see :py:meth:`normalize_points`) and the limiting density :math:`2r 1_{[0, 1]}(r)` of the radii is displayed

        :type normalization:
            bool, default ``True``

        .. seealso::

            - :py:meth:`normalize_points`
            - :py:meth:`sample_full_model`
            - :ref:`Full matrix model <ginibre_full_matrix_model>` associated to the Ginibre ensemble ensemble
        """

        self.__display_and_normalization('hist', normalization)
